pub mod registry {
    tonic::include_proto!("bky.registry.api");
}

use futures::Stream;
use registry::{PullCreateReq, PullCreateResp, PushCreateReq, PushCreateResp};
use std::{error::Error, io::ErrorKind, net::ToSocketAddrs, pin::Pin, result, time::Duration};
use tokio::sync::mpsc;
use tokio_stream::{wrappers::ReceiverStream, StreamExt};
use tonic::{transport::Server, Request, Response, Status, Streaming};

type PushCreateResult<T> = Result<Response<T>, Status>;
#[derive(Debug)]
pub struct RegistryServer {}

type RegistryResult<T> = Result<Response<T>, Status>;
type ResponseStream = Pin<Box<dyn Stream<Item = Result<PullCreateResp, Status>> + Send>>;

fn match_for_io_error(err_status: &Status) -> Option<&std::io::Error> {
    let mut err: &(dyn Error + 'static) = err_status;

    loop {
        if let Some(io_err) = err.downcast_ref::<std::io::Error>() {
            return Some(io_err);
        }

        // h2::Error do not expose std::io::Error with `source()`
        // https://github.com/hyperium/h2/pull/462
        if let Some(h2_err) = err.downcast_ref::<h2::Error>() {
            if let Some(io_err) = h2_err.get_io() {
                return Some(io_err);
            }
        }

        err = match err.source() {
            Some(err) => err,
            None => return None,
        };
    }
}
#[tonic::async_trait]
impl registry::metadata_ops_server::MetadataOps for RegistryServer {
    type CreateByPullStream = ResponseStream;
    async fn create_by_pull(
        &self,
        req: Request<Streaming<PullCreateReq>>,
    ) -> RegistryResult<Self::CreateByPullStream> {
        println!("create_by_pull ");
        println!("req={:?}", req);
        let mut in_stream = req.into_inner();
        println!("in_stream={:?}", in_stream);
        let (tx, rx) = mpsc::channel(128);
        tokio::spawn(async move {
            println!("tokio spawn..");
            while let Some(result) = in_stream.next().await {
                println!("result = {:?}", result);
                match result {
                    Ok(v) => tx
                        .send(Ok(PullCreateResp {
                            t_code: v.t_code,
                            ret_code: 0,
                        }))
                        .await
                        .expect("working rx"),
                    Err(err) => {
                        if let Some(io_err) = match_for_io_error(&err) {
                            if io_err.kind() == ErrorKind::BrokenPipe {
                                eprintln!("\tclient disconnected: broken pipe");
                                break;
                            }
                        }
                        match tx.send(Err(err)).await {
                            Ok(_) => (),
                            Err(_err) => break,
                        }
                    }
                }
                println!("\tstream ended");
            }
        });
        let out_stream = ReceiverStream::new(rx);
        Ok(Response::new(Box::pin(out_stream)))
    }

    //======================================
    type CreateByPushStream = Pin<Box<dyn Stream<Item = Result<PushCreateResp, Status>> + Send>>;
    async fn create_by_push(
        &self,
        req: Request<Streaming<PushCreateReq>>,
    ) -> RegistryResult<Self::CreateByPushStream> {
        println!("create_by_pull ");
        println!("req={:?}", req);
        let mut in_stream = req.into_inner();
        println!("in_stream={:?}", in_stream);
        let (tx, rx) = mpsc::channel(128);
        tokio::spawn(async move {
            println!("tokio spawn..");
            while let Some(result) = in_stream.next().await {
                println!("result = {:?}", result);
                match result {
                    Ok(v) => tx
                        .send(Ok(PushCreateResp {
                            tcode: v.tcode,
                            ret_code: 0,
                        }))
                        .await
                        .expect("working rx"),
                    Err(err) => {
                        if let Some(io_err) = match_for_io_error(&err) {
                            if io_err.kind() == ErrorKind::BrokenPipe {
                                eprintln!("\tclient disconnected: broken pipe");
                                break;
                            }
                        }
                        match tx.send(Err(err)).await {
                            Ok(_) => (),
                            Err(_err) => break,
                        }
                    }
                }
                println!("\tstream ended");
            }
        });
        let out_stream = ReceiverStream::new(rx);
        Ok(Response::new(Box::pin(out_stream)))
    }
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    println!("main...");
    let server = RegistryServer {};
    Server::builder()
        .add_service(registry::metadata_ops_server::MetadataOpsServer::new(
            server,
        ))
        .serve(
            "192.168.110.183:50050"
                .to_socket_addrs()
                .unwrap()
                .next()
                .unwrap(),
        )
        .await
        .unwrap();
    Ok(())
}
